/*
 * Test Flash block driver
 */

#define __KERNEL__
#define MODULE

#include <linux/config.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/ioctl.h>
#include <linux/blkpg.h>
#include <linux/fs.h>
#include <linux/mm.h>
#include <linux/delay.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
#include <asm/delay.h>
#include <asm/arch/hardware.h>

#include <linux/blk.h>

#include "myflash.h"

/* Flash mapping start */
#define FLASH_START		0xe8000000

static int flash_sectsize = FLASH_SECTSIZE;

#define WRITE_VERIFY
#define FLASH_DEBUG_CFI
#define FLASH_VERBOSE
#define FLASH_DEBUG

#ifdef SIMPUTER
#define FLASH_CMD_READ		( 0x00FF )
#define FLASH_CMD_CLEAR		( 0x0050 )
#define FLASH_CMD_WR_BUFF	( 0x00e8 )
#else
#define FLASH_CMD_READ		_CMD( 0x00FF )
#define FLASH_CMD_CLEAR		_CMD( 0x0050 )
#define FLASH_CMD_WR_BUFF	_CMD( 0x00e8 )
#endif

#define FLASH_CMD_READ_ID_CODES _CMD( 0x0090 )
#define FLASH_CMD_READ_QUERY	_CMD( 0x0098 )
#define FLASH_CMD_ERASE		_CMD( 0x0020 )
#define FLASH_CMD_CONFIRM	_CMD( 0x00D0 )
#define FLASH_CMD_WRITE		_CMD( 0x0040 )
#define FLASH_CMD_STATUS	_CMD( 0x0070 )

#define FLASH_STATUS_READY	_CMD( 0x0080 )

#define FLASH_WR_BUFF_SIZE	16

#define ERROR_VOLTAGE_RANGE	_CMD( 0x0008 )
#define ERROR_DEVICE_PROTECT    _CMD( 0x0002 )
#define ERROR_PROGRAMMING	_CMD( 0x0010 )

#define USEC		1000000		/* usec per sec */
#define POLL_TIME	10		/* usec */
#define POLL_LOOP_LIMIT	3*USEC/POLL_TIME  /* 3 sec */

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

static struct flash_cfi_info flash_info;

static void flash_retrieve_cfi_info(void)
{
	FLASH_t* flash_cmd_ptr = FLASH_PTR((char*)FLASH_START, 0);
	FLASH_t* flash_ptr = FLASH_PTR((char*)FLASH_START, 0 * sizeof(FLASH_t));
#define FLASH_LOBYTE(x) (flash_ptr[(x)] & 0x000000FFL)

#ifdef FLASH_DEBUG_CFI
	u8 testmanid;
	u8 testdevid;
#endif
	
	/* Clear the information structure */
	memset((void*)&flash_info, 0, sizeof(flash_info));
	
	/* Clear any status register info */
	*flash_cmd_ptr = FLASH_CMD_CLEAR;

	/* Put flash into READ ID code mode */
	*flash_cmd_ptr = FLASH_CMD_READ_ID_CODES;
	udelay(POLL_TIME);

	/* Get the manufacturer code */
	flash_info.manufacturer_code = FLASH_LOBYTE(0x00);

	flash_ptr = FLASH_PTR((char*)FLASH_START, 1 * sizeof(FLASH_t));
	/* Get the device code */
	flash_info.device_code = FLASH_LOBYTE(0x00);

	/* Put flash into READ QUERY mode */
	*flash_cmd_ptr = FLASH_CMD_READ_QUERY;
	udelay(POLL_TIME);

	flash_ptr = FLASH_PTR((char*)FLASH_START, 0 * sizeof(FLASH_t));

#ifdef FLASH_DEBUG_CFI
	testmanid = FLASH_LOBYTE(0x00);
	testdevid = FLASH_LOBYTE(0x01);
#endif
	
	/* Retrieve CFI Ident string */
	flash_info.cfi_ident[0] = (char)FLASH_LOBYTE(0x10);
	flash_info.cfi_ident[1] = (char)FLASH_LOBYTE(0x11);
	flash_info.cfi_ident[2] = (char)FLASH_LOBYTE(0x12);
	flash_info.cfi_ident[3] = (char)0;

	flash_info.is_cfi_supported = ((flash_info.cfi_ident[0] == 'Q') &&
				       (flash_info.cfi_ident[1] == 'R') &&
				       (flash_info.cfi_ident[2] == 'Y'));
		
	if (flash_info.is_cfi_supported != 0) {
		/* Retrieve CFI Ident extended command set info */
		flash_info.primary_vendor_command_code = FLASH_LOBYTE(0x13);
		flash_info.primary_vendor_command_code |=
			(FLASH_LOBYTE(0x14) << 8);

		flash_info.primary_extended_query_table = 
			FLASH_LOBYTE(0x15);
		flash_info.primary_extended_query_table |=
			(FLASH_LOBYTE(0x16) << 8);
		
		flash_info.alternate_vendor_command_code = 
			FLASH_LOBYTE(0x17);
		flash_info.alternate_vendor_command_code |=
			(FLASH_LOBYTE(0x18) << 8);
		
		flash_info.alternate_extended_query_table = 
			FLASH_LOBYTE(0x19);
		flash_info.alternate_extended_query_table |=
			(FLASH_LOBYTE(0x1A) << 8);
		
		/* Retrieve System Interface Information */
		flash_info.vcc_min_program_erase = FLASH_LOBYTE(0x1B);
		flash_info.vcc_max_program_erase = FLASH_LOBYTE(0x1C);
		flash_info.vpp_min_program_erase = FLASH_LOBYTE(0x1D);
		flash_info.vpp_max_program_erase = FLASH_LOBYTE(0x1E);
		flash_info.typ_single_word_timeout = FLASH_LOBYTE(0x1F);
		flash_info.typ_max_buffer_write_timeout =
			FLASH_LOBYTE(0x20);
		flash_info.typ_block_erase_timeout = FLASH_LOBYTE(0x21);
		flash_info.typ_full_chip_erase_timeout = FLASH_LOBYTE(0x22);
		flash_info.max_word_program_timeout = FLASH_LOBYTE(0x23);
		flash_info.max_buffer_write_timeout = FLASH_LOBYTE(0x24);
		flash_info.max_block_erase_timeout = FLASH_LOBYTE(0x25);
		flash_info.max_chip_erase_timeout = FLASH_LOBYTE(0x26);
		
		/* Retrieve Device Geometry Info */
		flash_info.device_size = FLASH_LOBYTE(0x27);
		flash_info.x8_async_ifc = FLASH_LOBYTE(0x28);
		flash_info.x16_async_ifc = FLASH_LOBYTE(0x29);
		flash_info.max_write_buffer_size = FLASH_LOBYTE(0x2A);
		flash_info.max_write_buffer_size |= 
			(FLASH_LOBYTE(0x2B) << 8);
		flash_info.num_erase_block_regions = FLASH_LOBYTE(0x2C);
		flash_info.num_erase_blocks_r1 = FLASH_LOBYTE(0x2D);
		flash_info.num_erase_blocks_r1 |= 
			(FLASH_LOBYTE(0x2E) << 8);
		flash_info.size_erase_blocks_r1 = FLASH_LOBYTE(0x2F);
		flash_info.size_erase_blocks_r1 |= 
			(FLASH_LOBYTE(0x30) << 8);
	}
		
#ifdef FLASH_DEBUG_CFI
	printk("Flash CFI Info: ****************************\n");
	printk("\tmanufacturer_code = %#04x (%#04x)\n", 
	       flash_info.manufacturer_code, testmanid);
	printk("\tdevice_code = %#04x (%#04x)\n", 
	       flash_info.device_code, testdevid);
	printk("\tcfi_ident = %s\n", flash_info.cfi_ident);
	printk("\tis_cfi_supported = %d\n", flash_info.is_cfi_supported);
	if (flash_info.is_cfi_supported != 0) {
		int tmpval1, tmpval2;

		/* Print rest of CFI Ident */
		printk("\tprimary_vendor_command_code = %#04x\n", 
		       flash_info.primary_vendor_command_code);
		printk("\tprimary_extended_query_table = %#04x\n",
		       flash_info.primary_extended_query_table);
		printk("\talternate_vendor_command_code = %#04x\n",
		       flash_info.alternate_vendor_command_code);
		printk("\talternate_extended_query_table = %#04x\n",
		       flash_info.alternate_extended_query_table);
		
		/* Print CFI Geometry Info */
		tmpval1 = 1 << flash_info.device_size;
		tmpval2 = tmpval1 / 1024 / 1024;
		printk("\tdevice_size = %#02x (%d bytes, %d MB, %d Mb)\n",
		       flash_info.device_size,
		       tmpval1, tmpval2, tmpval2 * 8);

		printk("\tx8_async_ifc = %#02x\n", 
		       flash_info.x8_async_ifc);
		printk("\tx16_async_ifc = %#02x\n", 
		       flash_info.x16_async_ifc);

		printk("\tmax_write_buffer_size = %d (%d bytes)\n", 
		       flash_info.max_write_buffer_size,
		       1 << flash_info.max_write_buffer_size);
		printk("\tnum_erase_block_regions = %d\n",
		       flash_info.num_erase_block_regions);
		printk("\tnum_erase_blocks_r1 = %#04x (%d blocks)\n",
		       flash_info.num_erase_blocks_r1,
		       flash_info.num_erase_blocks_r1);
		printk("\tsize_erase_blocks_r1 = %#04x (%d bytes)\n",
		       flash_info.size_erase_blocks_r1,
		       flash_info.size_erase_blocks_r1 * 256);

		/* System Interface Information */
		printk("\tvcc_min_program_erase = %d.%d V\n",
		       (flash_info.vcc_min_program_erase & 0xF0) >> 4,
		       flash_info.vcc_min_program_erase & 0x0F);
		printk("\tvcc_max_program_erase = %d.%d V\n",
		       (flash_info.vcc_max_program_erase & 0xF0) >> 4,
		       flash_info.vcc_max_program_erase & 0x0F);
		printk("\tvpp_min_program_erase = %d.%d V\n",
		       (flash_info.vpp_min_program_erase & 0xF0) >> 4,
		       flash_info.vpp_min_program_erase & 0x0F);
		printk("\tvpp_max_program_erase = %d.%d V\n",
		       (flash_info.vpp_max_program_erase & 0xF0) >> 4,
		       flash_info.vpp_max_program_erase & 0x0F);

		printk("\ttyp_single_word_timeout = %d (%d usec)\n",
		       flash_info.typ_single_word_timeout,
		       1 << flash_info.typ_single_word_timeout);
		printk("\ttyp_max_buffer_write_timeout = %d (%d usec)\n",
		       flash_info.typ_max_buffer_write_timeout,
		       1 << flash_info.typ_max_buffer_write_timeout);
		printk("\ttyp_block_erase_timeout = %d (%d msec)\n",
		       flash_info.typ_block_erase_timeout,
		       1 << flash_info.typ_block_erase_timeout);
		printk("\ttyp_full_chip_erase_timeout = %d (%d msec)\n",
		       flash_info.typ_full_chip_erase_timeout,
		       1 << flash_info.typ_full_chip_erase_timeout);

		printk("\tmax_word_program_timeout = %d (%d usec)\n",
		       flash_info.max_word_program_timeout,
		       ((1 << flash_info.max_word_program_timeout) *
			(1 << flash_info.typ_single_word_timeout)));
		printk("\tmax_buffer_write_timeout = %d (%d usec)\n",
		       flash_info.max_buffer_write_timeout,
		       ((1 << flash_info.max_buffer_write_timeout) *
			(1 << flash_info.typ_max_buffer_write_timeout)));
		printk("\tmax_block_erase_timeout = %d (%d msec)\n",
		       flash_info.max_block_erase_timeout,
		       ((1 << flash_info.max_block_erase_timeout) * 
			(1 << flash_info.typ_block_erase_timeout)));
		printk("\tmax_chip_erase_timeout = %d (%d msec)\n",
		       flash_info.max_chip_erase_timeout,
		       ((1 << flash_info.max_chip_erase_timeout) *
			(1 << flash_info.typ_full_chip_erase_timeout)));
	}
	
#endif /* FLASH_DEBUG */
	
	/* Clear any status register info */
	*flash_cmd_ptr = FLASH_CMD_CLEAR;

	/* Return flash to read mode */
	*flash_cmd_ptr = FLASH_CMD_READ;
}

int __init flash_mem_init(void)
{
	int		i;

	flash_retrieve_cfi_info();
	
	if (flash_info.is_cfi_supported != 0) {
		/* Set true flash sector size */
		flash_sectsize = SECTSIZE_FROM_CFI( FLASH_CHIPS_PER_BUS,
						    flash_info.size_erase_blocks_r1);

		/* find appropriate flash */
		for (i = 0; flashInfoList[i] != NULL; ++i) {
			if ((flash_info.manufacturer_code == 
			     flashInfoList[i]->manufacturer_code) &&
			    (flash_info.device_code == 
			     flashInfoList[i]->device_code)) {
				/* Found flash */
				printk("FLASH Driver: using %s devices\n",
				       flashInfoList[i]->deviceName);
				
				memcpy(flash_length,
				       flashInfoList[i]->flash_length,
				       sizeof(flash_length));
				break;
			}
		}

		/* verify flash */
		if (flashInfoList[i] == NULL) {
			memcpy((void*)flash_length,
			       (void*)flash_length_default,
			       sizeof(flash_length_default));
			
			printk(KERN_WARNING
			       "FLASH Driver: Unknown flash type (manufacturer = %#04x, device_code = %#04x!\n",
			       flash_info.manufacturer_code,
			       flash_info.device_code);
		}
			
		printk("FLASH Driver: flash_sectsize = %d bytes (%d KB)\n",
		       flash_sectsize, flash_sectsize / 1024);
	}

	printk("Test FLASH driver initialized\n" );

	return(0);
}

#if 0
static int write_sector( char *flash_addr, char *data_addr )
{
	volatile FLASH_t *flash_ptr;
	FLASH_t *data_ptr;
	unsigned long i;
	int error = FALSE;

#ifdef FLASH_VERBOSE
	printk( "Writing sector %p with data at %p\n", 
		flash_addr, data_addr);
	/* Vivek */
	printk(KERN_CRIT"Before:\n");
	for(i=0;i<6;i++)
		printk(KERN_CRIT"i=%d:f_addr=%p:f_data=%c:s_data=%c\n",
			i, flash_addr + i, *(flash_addr + i), *(data_addr + i));
#endif

	WP_VPP_ON();
	data_ptr = (FLASH_t *)data_addr;

	flash_ptr = FLASH_PTR(flash_addr, 0);
	*flash_ptr = FLASH_CMD_CLEAR;

	i = 0;
	while( i < flash_sectsize ) {
#ifdef USE_WRITE_BUFFER
		int pollmax = 1<<20;
		int cnt = FLASH_WR_BUFF_SIZE;

		flash_ptr = FLASH_PTR(flash_addr, i);
		do {
			*flash_ptr = FLASH_CMD_WR_BUFF;
			pollmax--;
		} while( pollmax && (*flash_ptr & FLASH_STATUS_READY) != FLASH_STATUS_READY );
		error = !flash_status_check( flash_ptr );
		if( error )  break;

		*flash_ptr = _CMD(cnt-1);
		while( cnt-- ) {
			flash_ptr = FLASH_PTR(flash_addr, i);
			i += sizeof(FLASH_t);
			*flash_ptr = *data_ptr++;
		}
		*flash_ptr = FLASH_CMD_CONFIRM;
		error = !flash_status_check( flash_ptr );
		if( error )  break;
#else
		flash_ptr = FLASH_PTR(flash_addr, i);
		i += sizeof(FLASH_t);
		*flash_ptr = FLASH_CMD_WRITE;
		*flash_ptr = *data_ptr++;
		error = !flash_status_check( flash_ptr );
		if( error )  break;
#endif
	}

	flash_ptr = FLASH_PTR(flash_addr, 0);
	*flash_ptr = FLASH_CMD_READ;
	WP_VPP_OFF();

#if 0
	/* Vivek */
	printk(KERN_CRIT"changing...\n");
	for(i=0;i<6;i+=4)
		*(unsigned int *)(flash_addr + i) = *(unsigned int *)
							(data_addr + i);
#endif

	/* Vivek */
	printk(KERN_CRIT"After:\n");
	for(i=0;i<6;i++)
		printk(KERN_CRIT"i=%d:f_addr=%p:f_data=%c\n",
			i, flash_addr + i, *(flash_addr + i));

	if( error ) return FALSE;

#ifdef WRITE_VERIFY
	return verify_data(flash_addr, data_addr, flash_sectsize);
#else
	return TRUE;
#endif
}
#endif

static int __init flash_init_module(void)
{
    int timeout,result,i;
	volatile FLASH_t *flash_ptr, *flash_read_addr, *flash_write_addr;
    FLASH_t x;

    //MSC0 = 0xFFF8FFFC;
    result = flash_mem_init();

    if(result)
        return result;

    printk(KERN_CRIT"Flash: testing...\n");
    flash_ptr = FLASH_PTR( (char *)FLASH_START, 0 );
    
    WP_VPP_ON();
    /* Start write */ 
    *flash_ptr = _CMD (0x0060); /* Clear block lock */
    *flash_ptr = FLASH_CMD_CONFIRM;
    *flash_ptr = FLASH_CMD_ERASE;
    *flash_ptr = FLASH_CMD_CONFIRM;
    *flash_ptr = FLASH_CMD_STATUS;
    x = *flash_ptr;

    timeout = POLL_LOOP_LIMIT;
    for (x = *flash_ptr; (x & FLASH_STATUS_READY) != FLASH_STATUS_READY; 
            x = *flash_ptr)
    {
        if (timeout -- < 0 )
            break;
        udelay(POLL_TIME);
    }

    printk(KERN_CRIT"Flash driver: (b4 reading & writing) status x = %x\n", x);
    if ( x != FLASH_STATUS_READY )
        return 1;

    for(i=0; i<20; i += sizeof(FLASH_t) )
    {
        flash_read_addr = FLASH_PTR((char *)FLASH_START, i);
        *flash_read_addr = FLASH_CMD_CLEAR;
        *flash_read_addr = 0xffffffff;
        x = *flash_read_addr;
        
        printk(KERN_CRIT"Flash: (reading b4 writing)i=%d,addr=%x,x=%x\n",i,
                            flash_read_addr,x);
    }

    for(i=0; i<20; i += sizeof(FLASH_t))
    {
        flash_write_addr = FLASH_PTR((char *)FLASH_START, i);
        *flash_write_addr = FLASH_CMD_WRITE;
        *flash_write_addr = i;

        timeout = POLL_LOOP_LIMIT;
        *flash_write_addr = FLASH_CMD_STATUS;
        for (x = *flash_write_addr; (x & FLASH_STATUS_READY) != FLASH_STATUS_READY; 
                x = *flash_write_addr)
        {
            if (timeout -- < 0 )
                break;
            udelay(POLL_TIME);
        }
        printk(KERN_CRIT"Flash: (writing) status i=%d,addr=%x,x=%x\n",i,
                            flash_write_addr,x);
        if ( x != FLASH_STATUS_READY  && timeout < 0 )
            return 1;
    }

    for(i=0; i<20; i +=sizeof(FLASH_t))
    {
        flash_read_addr = FLASH_PTR((char *)FLASH_START, i);
        *flash_read_addr = FLASH_CMD_CLEAR;
        *flash_read_addr = 0xffffffff;
        x = *flash_read_addr;
        
        printk(KERN_CRIT"Flash:(reading) i=%d,addr=%x,x=%x\n",i,
                            flash_read_addr,x);
    }
    WP_VPP_OFF();

    return result;
}


static void __exit flash_cleanup_module(void)
{
	printk(KERN_CRIT"Flash driver: Unloading...\n");
}


/* The default module stuff */
module_init(flash_init_module);
module_exit(flash_cleanup_module);

MODULE_AUTHOR("Vivek");
MODULE_DESCRIPTION("Test Flash memory driver");

EXPORT_NO_SYMBOLS;
